global float camRot;
global float camX;
global float camY;
global float camZ;

global float gravity;

[0,2,7,11,12] @=> global int scale[];

public class Bubble extends GGen {
    Keyboard keyboard;
    spork ~ keyboard.listener();

    VectorMath m;

    Gain lfoG => Gain g => NRev rev => Pan2 pan => dac;
    

    SinOsc sin => Envelope env => lfoG;
    TriOsc tri => env;


    Pop popSound => g;
    popSound.gain(5);
    
    60 + 12*Math.random2(0,2) + scale[Math.random2(0, scale.size()-1)] => int midi_number;
    Std.mtof( midi_number ) => sin.freq;
    Std.mtof( midi_number + 12 ) => tri.freq;

    0.7 => sin.gain;
    0.1 => tri.gain;
    // <<< "freq: " , sin.freq() >>>;
    g.gain(0.001);
    env.duration(0.5::second);

    Math.random2f(2,10) => float lfo;

    fun void LFOGain() {
        while (true) {
            GG.nextFrame() => now;
            lfoG.gain(0.5 + 0.3*Math.sin(lfo*(now/second)));
        }
    } spork ~ LFOGain();


    spork ~ play();

    fun void play() {
        env.keyOn();
    }
    fun void stop() {
        env.keyOff();
    }

    float distance;
    vec3 currentPos;
    float angle;

    fun void updateDistance() {
        while (true) {
            GG.nextFrame() => now;
            this.posWorld() => currentPos;
            m.norm(currentPos) => distance;

            if (distance > 500) {
                this.detach();
            }

            m.normalize(currentPos) => currentPos;

            m.angleBetween( @(currentPos.x, 0, currentPos.z), @(camX, 0, camZ) ) => angle;

        }
    } spork ~ updateDistance();

    int sign;
    fun void updateParams() {
        while (true) {
            GG.nextFrame() => now;
            Math.min(0.005, 1 / (distance*distance)) => g.gain;
            0.1 * Math.map2(distance, 0, 500, 0, 1) => rev.mix;

            m.cross( @(currentPos.x, 0, currentPos.z), @(camX, 0, camZ) ) => vec3 cross;
            if (cross.y > 0) {
                1 => sign;
            } else {
                -1 => sign;
            }


            // Math.map2(angle, -Math.PI, Math.PI, -1, 1) => pan.pan;
            sign * Math.sin(angle) => pan.pan;
            // <<< "pan: ", pan.pan() >>>;

        }
    } spork ~ updateParams();









    
    0.75*GG.dt() => float dt;
    0.005 => float timeFactor; 


    Texture.load(me.dir() + "textures/swirl.png") @=> Texture swirl;

    PBRMaterial bubblemat; 
    // bubblemat.envmapBlend(PhongMaterial.EnvmapBlend_Add);
    // bubblemat.envmapBlend(2);
    // bubblemat.envmapIntensity(1.0);
    // bubblemat.refractionRatio(.95);
    // bubblemat.envmapMethod(2);
    bubblemat.colorMap(swirl);
    bubblemat.emissive( Color.BLUE );
    bubblemat.color( @(1,1,1) * 7.5);
    bubblemat.aoFactor(5);

    0.1 => float radius;

    SphereGeometry bubbleGeo(
        radius, // float radius, 
        32, // int widthSeg, 
        32, // int heightSeg, 
        0, // float phiStart, 
        2*Math.PI, // float phiLength, 
        0, // float thetaStart, 
        Math.PI // float thetaLength
    );

    GMesh bubble(bubbleGeo, bubblemat) --> this;
    bubble.rot( @(1,1,1)*Math.random2f(0,2*Math.PI) );
    bubble.sca(Math.random2f(0.9, 1.1));
    bubble.sca() => vec3 initialScale;

    // GPointLight light --> bubble;
    // light.intensity(0);


    @(Math.random2f(0,10), Math.random2f(0,10), Math.random2f(0,10)) => vec3 squishFactor;

    fun void squish() {
        while (true) {
            bubble.sca( initialScale + 
                        0.25 * @( Math.sin(squishFactor.x * (now/second)) ,
                                   Math.sin(squishFactor.y * (now/second)) ,
                                   Math.sin(squishFactor.z * (now/second)) ) );
            GG.nextFrame() => now;  
            if (this.posWorld().y < 0 || distance > 300) {
                break;
            }
        }
    } spork ~ squish();



    
    // constants
    float tMain;
    vec3 accTermMain;
    float tFactorMain;

     
    vec3 initialVelocityMain;

    fun void setInitialVelocity(vec3 blowDirection, float blowStrength) {
        // add random directions
        blowDirection + @(Math.random2f(-1,1), 0, Math.random2f(-1,1)) => initialVelocityMain;

        //normalize
        m.normalize(initialVelocityMain) => initialVelocityMain;
        // make the y component much smaller
        0.5 * initialVelocityMain.y => initialVelocityMain.y;
        
        // scale by blowStrength
        blowStrength * initialVelocityMain => initialVelocityMain;
    }


    vec3 previousPos; // To store the position from the previous frame
    vec3 velocity;
    vec3 drag;
    0.01 => float dragScaling;
    float speed;
    float previousSpeed;

    fun void Physics() {

        // update time
        tMain + dt => tMain;
        tMain * timeFactor => tFactorMain;


        @(0,  0.5 * gravity * Math.fabs(Math.pow(tFactorMain, 2.5)), 0) => accTermMain;

        this.pos() + initialVelocityMain * tFactorMain + accTermMain => this.pos;

        0.999*initialVelocityMain => initialVelocityMain;

    }


    // Math.random2f(0.1,3) => float flickerRate;
    // Math.random2f(500,2500) => float howLongToWait;
    // fun void lightUp() {
    //     while (true) {
    //         if (tMain > howLongToWait*dt) {
    //             <<< "lighting up" >>>;
    //             0.5 - 0.25*Math.cos( flickerRate*(now/second)) => light.intensity;
    //         }
    //         GG.nextFrame() => now;
    //     }
    // }   

    // // only some of the bubbles light up
    // if (Math.random2f(0,1) < 0.3) {
    //     spork ~ lightUp();
    // }






    SphereGeometry burstGeo(
        radius / 15.0, // float radius, 
        8, // int widthSeg, 
        8, // int heightSeg, 
        0, // float phiStart, 
        2*Math.PI, // float phiLength, 
        0, // float thetaStart, 
        Math.PI // float thetaLength
    );

    fun void pop(int duration) {
        spork ~ stop();
        // <<< "pop!" >>>;

        // play pop sound
        spork ~ popSound.play();


        30 => int n;
        n*n => int numPoints;  

        GMesh points[numPoints];
        vec3 initialVelocities[numPoints];


        float theta;
        float phi;

        for (0 => int j; j < n; j++) {

            2*Math.PI * (j / n$float) => theta;

            for (0 => int i; i < n; i++) {

                Math.PI * (i / n$float) => phi;

                points[i + (j*n) ].pos( radius * 
                                        @(Math.sin(phi) * Math.cos(theta),
                                          Math.sin(phi) * Math.sin(theta),
                                          Math.cos(phi) ) );
                points[i + (j*n) ].sca( 0.5 );
                points[i + (j*n) ].mesh( burstGeo, bubblemat );
                points[i + (j*n) ] --> this;

                // <<< "index: " , i + (j*n) >>>;

                @(  Math.random2f(-0.4, 0.4), 
                    Math.random2f( 0, 0.5), 
                    Math.random2f(-0.4, 0.4) ) => initialVelocities[i + (j*n)];
            }
        }
        bubble.detach();



        float t;
        vec3 gravityTerm;
        float tFactor;

        vec3 currentPos;

        for (int n; n < duration; n++) {
            GG.nextFrame() => now;

            t + dt => t;

            t * timeFactor => tFactor;

            @(0, -0.5 * gravity * Math.pow(tFactor, 2), 0) => gravityTerm;
            
            for (int i; i < numPoints; i++) {

                points[i].posWorld() => currentPos;

                if (Math.random2f(0,1) < 0.1) {
                    points[i].detach();
                }

                currentPos
                + initialVelocities[i] * tFactor
                + gravityTerm
                    => points[i].posWorld;


            }

        }

        for (auto p : points) {
            p.detach();
        }
    }

    int alreadyPopped;

    fun void maybePop() {
        while(true) {
            
            GG.nextFrame() => now;

            if (alreadyPopped) {
                break;
            }

            if (Math.random2f(0,1) < 0.0005 && !alreadyPopped) {
                true => alreadyPopped;
                // <<< " trying to pop " >>>;
                pop(Math.random2(20,40));
                this.detach();
                break;
            }
        }
    } spork ~ maybePop();


    fun void forcePop() {
        while (true) {
            GG.nextFrame() => now;
            if (keyboard.P) {
                if (!alreadyPopped) {
                    true => alreadyPopped;
                    pop(Math.random2(20,100));
                    this.detach();

                    break;
                }
            }
        }
    } spork ~ forcePop();


    int detached;
    fun void detachFromWand() {
        spork ~ popSound.play();
        true => detached;
    }

    fun void update(float dt) {
        if (this.posWorld().y > 0 && detached) {
            Physics();
        }
    }
}



